# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from abc import abstractmethod
from dataclasses import InitVar as dataclass_InitVar
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Self,
    Sequence,
    Type,
    cast,
    get_args,
    get_origin,
)

from deprecated import deprecated
from more_itertools import partition

from faebryk.core.cpp import Node as CNode
from faebryk.core.graphinterface import (
    GraphInterface,
)
from faebryk.core.link import LinkNamedParent, LinkSibling
from faebryk.libs.exceptions import UserException
from faebryk.libs.util import (
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    Tree,
    cast_assert,
    find,
    in_debug_session,
    not_none,
    once,
    post_init_decorator,
    times,
    zip_dicts_by_key,
)

if TYPE_CHECKING:
    from faebryk.core.solver.solver import Solver
    from faebryk.core.trait import Trait, TraitImpl

logger = logging.getLogger(__name__)


# TODO: should this be a FaebrykException?
# TODO: should this include node and field information?
class FieldError(Exception):
    pass


class FieldExistsError(FieldError):
    pass


class FieldContainerError(FieldError):
    pass


def list_field[T: Node](n: int, if_type: Callable[[], T]) -> list[T]:
    return d_field(lambda: times(n, if_type))


class fab_field:
    pass


class constructed_field[T: "Node", O: "Node"](property, fab_field):
    """
    Field which is constructed after the node is created.
    The constructor gets one argument: the node instance.

    The constructor should return the constructed faebryk object or None.
    If a faebryk object is returned, it will be added to the node.
    """

    @abstractmethod
    def __construct__(self, obj: T) -> O | None:
        pass


class rt_field[T, O](constructed_field):
    """
    rt_fields (runtime_fields) are the last fields excecuted before the
    __preinit__ and __postinit__ functions are called.
    It gives the function passed to it access to the node instance.
    This is useful to do construction that depends on parameters passed by __init__.
    """

    def __init__(self, fget: Callable[[T], O]) -> None:
        super().__init__()
        self.func = fget
        self.lookup: dict[T, O] = {}

    def __construct__(self, obj: T):
        constructed = self.func(obj)
        # TODO find a better way for this
        # in python 3.13 name support
        self.lookup[obj] = constructed

        return constructed

    def __get__(self, instance: T, owner: type | None = None) -> Any:
        return self.lookup[instance]


class _d_field[T](fab_field):
    def __init__(self, default_factory: Callable[[], T]) -> None:
        self.default_factory = default_factory

    def __repr__(self) -> str:
        return f"{super().__repr__()}{self.default_factory=})"


def d_field[T](default_factory: Callable[[], T]) -> T:
    return _d_field(default_factory)  # type: ignore


def f_field[T, **P](con: Callable[P, T]) -> Callable[P, T]:
    assert isinstance(con, type)

    def _(*args: P.args, **kwargs: P.kwargs) -> Callable[[], T]:
        def __() -> T:
            return con(*args, **kwargs)

        return _d_field(__)  # type: ignore

    return _  # type: ignore


def list_f_field[T, **P](n: int, con: Callable[P, T]) -> Callable[P, list[T]]:
    assert isinstance(con, type)

    def _(*args: P.args, **kwargs: P.kwargs) -> Callable[[], list[T]]:
        def __() -> list[T]:
            return [con(*args, **kwargs) for _ in range(n)]

        return _d_field(__)  # type: ignore

    return _  # type: ignore


class NodeException(UserException):
    def __init__(self, node: "Node", *args: object) -> None:
        super().__init__(*args)
        self.node = node


class FieldConstructionError(UserException):
    def __init__(self, node: "Node", field: str, *args: object) -> None:
        super().__init__(*args)
        self.node = node
        self.field = field


class NodeAlreadyBound(NodeException):
    def __init__(self, node: "Node", other: "Node", *args: object) -> None:
        super().__init__(
            node,
            *args,
            f"Node {other} already bound to"
            f" {other.get_parent()}, can't bind to {node}",
        )


class NodeNoParent(NodeException): ...


class InitVar(dataclass_InitVar):
    """
    This is a type-marker which instructs the Node constructor to ignore the field.

    Inspired by dataclasses.InitVar, which it inherits from.
    """


# -----------------------------------------------------------------------------


@post_init_decorator
class Node(CNode):
    runtime_anon: list["Node"]
    runtime: dict[str, "Node"]
    specialized_: list["Node"]

    _init: bool = False
    _mro: list[type] = []
    _mro_ids: set[int] = set()

    class _Skipped(Exception):
        pass

    def add[T: Node | GraphInterface](
        self,
        obj: T,
        name: str | None = None,
        container: Sequence | dict[str, Any] | None = None,
    ) -> T:
        assert obj is not None

        if container is None:
            if name:
                container = self.runtime
            else:
                container = self.runtime_anon

        try:
            container_name = find(vars(self).items(), lambda x: x[1] is container)[0]
        except KeyErrorNotFound:
            raise FieldContainerError("Container not in fields")

        if name:
            if not isinstance(container, dict):
                raise FieldContainerError(f"Expected dict got {type(container)}")
            if name in container:
                raise FieldExistsError(name)
            # TODO consider setting name for non runtime container
            # if container is not self.runtime:
            #   name = f"{container_name}[{name}]"
            pass
        else:
            if not isinstance(container, list):
                raise FieldContainerError(f"Expected list got {type(container)}")
            name = f"{container_name}[{len(container)}]"

        try:
            if isinstance(obj, GraphInterface):
                self._handle_add_gif(name, obj)
            else:
                self._handle_add_node(name, obj)
        except Node._Skipped:
            return obj

        # add to container
        if isinstance(container, dict):
            container[name] = obj
        else:
            container.append(obj)

        return obj

    def add_to_container[T: Node](
        self,
        n: int,
        factory: Callable[[], T],
        container: Sequence["Node"] | None = None,
    ):
        if container is None:
            container = self.runtime_anon

        constr = [factory() for _ in range(n)]
        for obj in constr:
            self.add(obj, container=container)
        return constr

    @classmethod
    @once
    def __faebryk_fields__(cls) -> tuple[dict[str, Any], dict[str, Any]]:
        def all_vars(cls):
            return {k: v for c in reversed(cls.__mro__) for k, v in vars(c).items()}

        def all_anno(cls):
            return {
                k: v
                for c in reversed(cls.__mro__)
                if hasattr(c, "__annotations__")
                for k, v in c.__annotations__.items()
            }

        LL_Types = (Node, GraphInterface)

        vars_ = {
            name: obj
            for name, obj in all_vars(cls).items()
            # private fields are always ignored
            if not name.startswith("_")
            # only consider fab_fields
            and isinstance(obj, fab_field)
        }
        annos = {
            name: obj
            for name, obj in all_anno(cls).items()
            # private fields are always ignored
            if not name.startswith("_")
            # explicitly ignore InitVars
            and not isinstance(obj, InitVar)
            # variables take precedence over annos
            and name not in vars_
        }

        # ensure no field annotations are a property
        # If properties are constructed as instance fields, their
        # getters and setters aren't called when assigning to them.
        #
        # This means we won't actually construct the underlying graph properly.
        # It's pretty insidious because it's completely non-obvious that we're
        # missing these graph connections.
        # TODO: make this an exception group instead
        for name, obj in annos.items():
            if (origin := get_origin(obj)) is not None:
                # you can't truly subclass properties because they're a descriptor
                # type, so instead we check if the origin is a property via our fields
                if issubclass(origin, constructed_field):
                    raise FieldError(
                        f"{name} is a property, which cannot be created from a field "
                        "annotation. Please instantiate the field directly."
                    )

        # FIXME: something's fucked up in the new version of this,
        # but I can't for the life of me figure out what
        clsfields_unf_new = dict(chain(annos.items(), vars_.items()))
        clsfields_unf_old = {
            name: obj
            for name, obj in chain(
                (
                    (name, obj)
                    for name, obj in all_anno(cls).items()
                    if not isinstance(obj, InitVar)
                ),
                (
                    (name, f)
                    for name, f in all_vars(cls).items()
                    if isinstance(f, fab_field)
                ),
            )
            if not name.startswith("_")
        }
        assert clsfields_unf_old == clsfields_unf_new
        clsfields_unf = clsfields_unf_old

        def is_node_field(obj):
            def is_genalias_node(obj):
                origin = get_origin(obj)
                assert origin is not None

                if issubclass(origin, LL_Types):
                    return True

                if issubclass(origin, (list, dict)):
                    arg = get_args(obj)[-1]
                    return is_node_field(arg)

            if isinstance(obj, LL_Types):
                raise FieldError("Node instances not allowed")

            if isinstance(obj, str):
                return obj in [L.__name__ for L in LL_Types]

            if isinstance(obj, type):
                return issubclass(obj, LL_Types)

            if isinstance(obj, _d_field):
                return True

            if get_origin(obj):
                return is_genalias_node(obj)

            if isinstance(obj, constructed_field):
                return True

            return False

        nonfabfields, fabfields = partition(
            lambda x: is_node_field(x[1]), clsfields_unf.items()
        )

        return dict(fabfields), dict(nonfabfields)

    def _setup_fields(self):
        clsfields, _ = self.__faebryk_fields__()
        LL_Types = (Node, GraphInterface)

        # for name, obj in clsfields_unf.items():
        #    if isinstance(obj, _d_field):
        #        obj = obj.type
        #    filtered = name not in clsfields
        #    filtered_str = "   FILTERED" if filtered else ""
        #    print(
        #        f"{cls.__qualname__+"."+name+filtered_str:<60} = {str(obj):<70} "
        # "| {type(obj)}"
        #    )

        added_objects: dict[str, Node | GraphInterface] = {}
        objects: dict[str, Node | GraphInterface] = {}

        def handle_add(name, obj):
            del objects[name]
            try:
                if isinstance(obj, GraphInterface):
                    self._handle_add_gif(name, obj)
                elif isinstance(obj, Node):
                    self._handle_add_node(name, obj)
                else:
                    raise TypeError(
                        f"Cannot handle adding field {name=} of type {type(obj)}"
                    )
            except Node._Skipped:
                return
            added_objects[name] = obj

        def append(name, inst):
            if isinstance(inst, LL_Types):
                objects[name] = inst
            elif isinstance(inst, list):
                for i, obj in enumerate(inst):
                    assert obj is not None
                    objects[f"{name}[{i}]"] = obj
            elif isinstance(inst, dict):
                for k, obj in inst.items():
                    objects[f"{name}[{k}]"] = obj

            return inst

        def _setup_field(name, obj):
            if isinstance(obj, str):
                raise NotImplementedError()

            if (origin := get_origin(obj)) is not None:
                if isinstance(origin, type):
                    setattr(self, name, append(name, origin()))
                    return
                raise NotImplementedError(origin)

            if isinstance(obj, _d_field):
                setattr(self, name, append(name, obj.default_factory()))
                return

            if isinstance(obj, type):
                setattr(self, name, append(name, obj()))
                return

            if isinstance(obj, constructed_field):
                if (constructed := obj.__construct__(self)) is not None:
                    append(name, constructed)
                return

            raise NotImplementedError()

        def setup_field(name, obj):
            try:
                _setup_field(name, obj)
            except Exception as e:
                # this is a bit of a hack to provide complete context to debuggers
                # for underlying field construction errors
                if in_debug_session():
                    raise
                raise FieldConstructionError(
                    self,
                    name,
                    f'An exception occurred while constructing field "{name}"',
                ) from e

        nonrt, rt = partition(
            lambda x: isinstance(x[1], constructed_field), clsfields.items()
        )
        for name, obj in nonrt:
            setup_field(name, obj)

        for name, obj in list(objects.items()):
            handle_add(name, obj)

        # rt fields depend on full self
        for name, obj in rt:
            setup_field(name, obj)

            for name, obj in list(objects.items()):
                handle_add(name, obj)

        return added_objects, clsfields

    def __new__(cls, *args, **kwargs):
        out = super().__new__(cls)
        return out

    def _setup(self, *args, **kwargs) -> None:
        assert not hasattr(self, "_setup_done")
        self._setup_done = False
        # Construct Fields
        _, _ = self._setup_fields()

        # Call 2-stage constructors

        if self._init:
            for f_name in ("__preinit__", "__postinit__"):
                for base in reversed(type(self).mro()):
                    if f_name in base.__dict__:
                        f = getattr(base, f_name)
                        f(self)
        self._setup_done = True

    def __hash__(self):
        return id(self)

    def __init__(self):
        super().__init__()
        CNode.transfer_ownership(self)
        assert not hasattr(self, "_called_init")
        self._called_init = True

    def __preinit__(self, *args, **kwargs) -> None: ...

    def __postinit__(self, *args, **kwargs) -> None: ...

    def __post_init__(self, *args, **kwargs):
        if not getattr(self, "_called_init", False):
            raise Exception(
                "Node constructor hasn't been called."
                "Did you forget to call super().__init__()?"
                f"{type(self).__qualname__}"
            )
        self._setup(*args, **kwargs)

    def __init_subclass__(cls, *, init: bool = True) -> None:
        cls._init = init
        post_init_decorator(cls)
        Node_mro = CNode.mro()
        cls_mro = cls.mro()
        cls._mro = cls_mro[: -len(Node_mro)]
        cls._mro_ids = {id(c) for c in cls._mro}

        # check if accidentally added a node instance instead of field
        node_instances = [
            (name, f)
            for name, f in vars(cls).items()
            if isinstance(f, Node) and not name.startswith("_")
        ]
        if node_instances:
            raise FieldError(f"Node instances not allowed: {node_instances}")

    def _handle_add_gif(self, name: str, gif: GraphInterface):
        gif.node = self
        gif.name = name
        gif.connect(self.self_gif, LinkSibling())

    def _handle_add_node(self, name: str, node: "Node"):
        if node.get_parent():
            raise NodeAlreadyBound(self, node)

        from faebryk.core.trait import TraitImpl

        if TraitImpl.is_traitimpl(node):
            if self.has_trait(node.__trait__):
                if not node.handle_duplicate(
                    cast(TraitImpl, self.get_trait(node.__trait__)), self
                ):
                    raise Node._Skipped()

        node.parent.connect(self.children, LinkNamedParent(name))
        node._handle_added_to_parent()

    def _remove_child(self, node: "Node"):
        node.parent.disconnect_parent()

    def _handle_added_to_parent(self): ...

    def builder(self, op: Callable[[Self], Any]) -> Self:
        op(self)
        return self

    @property
    def _is_setup(self) -> bool:
        return getattr(self, "_setup_done", False)

    # printing -------------------------------------------------------------------------

    def __str__(self) -> str:
        return self.get_full_name()

    def pretty_params(self, solver: "Solver | None" = None) -> str:
        from faebryk.core.parameter import Parameter

        params = {
            not_none(p.get_parent())[1]: p
            for p in self.get_children(direct_only=True, types=Parameter)
        }
        params_str = "\n".join(
            f"{k}: {solver.inspect_get_known_supersets(v)
                    if solver else v}"
            for k, v in params.items()
        )

        return params_str

    def relative_address(self, root: "Node | None" = None) -> str:
        """Return the address from root to self"""
        if root is None:
            return self.get_full_name()

        root_name = root.get_full_name()
        self_name = self.get_full_name()
        if not self_name.startswith(root_name):
            raise ValueError(f"Root {root_name} is not an ancestor of {self_name}")

        return self_name.removeprefix(root_name + ".")

    # Trait stuff ----------------------------------------------------------------------

    @deprecated("Just use add")
    def add_trait[_TImpl: "TraitImpl"](self, trait: _TImpl) -> _TImpl:
        return self.add(trait)

    def _find_trait_impl[V: "Trait | TraitImpl"](
        self, trait: type[V], only_implemented: bool
    ) -> V | None:
        from faebryk.core.trait import (
            Trait,
            TraitImpl,
            TraitImplementationConfusedWithTrait,
        )

        if TraitImpl.is_traitimpl_type(trait):
            if not trait.__trait__.__decless_trait__:
                raise TraitImplementationConfusedWithTrait(
                    self, cast(type[Trait], trait)
                )
            trait = trait.__trait__

        out = self.get_children(
            direct_only=True,
            types=Trait,
            f_filter=lambda impl: trait.is_traitimpl(impl)
            and (cast(TraitImpl, impl).is_implemented() or not only_implemented),
        )

        if len(out) > 1:
            raise KeyErrorAmbiguous(duplicates=list(out))
        assert len(out) <= 1
        return cast_assert(trait, next(iter(out))) if out else None

    def del_trait(self, trait: type["Trait"]):
        impl = self._find_trait_impl(trait, only_implemented=False)
        if not impl:
            return
        self._remove_child(impl)

    def try_get_trait[V: "Trait | TraitImpl"](self, trait: Type[V]) -> V | None:
        return self._find_trait_impl(trait, only_implemented=True)

    def has_trait(self, trait: type["Trait | TraitImpl"]) -> bool:
        try:
            return self.try_get_trait(trait) is not None
        except KeyErrorAmbiguous:
            return True

    def get_trait[V: "Trait | TraitImpl"](self, trait: Type[V]) -> V:
        from faebryk.core.trait import Trait, TraitNotFound

        impl = self.try_get_trait(trait)
        if not impl:
            raise TraitNotFound(self, cast(type[Trait], trait))

        return cast_assert(trait, impl)

    # Graph stuff ----------------------------------------------------------------------
    def get_children[T: Node](
        self,
        direct_only: bool,
        types: type[T] | tuple[type[T], ...],
        include_root: bool = False,
        f_filter: Callable[[T], bool] | None = None,
        sort: bool = True,
    ) -> set[T]:
        return cast(
            set[T],
            set(
                super().get_children(
                    direct_only=direct_only,
                    types=types if isinstance(types, tuple) else (types,),
                    include_root=include_root,
                    f_filter=f_filter,  # type: ignore
                    sort=sort,
                )
            ),
        )

    def get_tree[T: Node](
        self,
        types: type[T] | tuple[type[T], ...],
        include_root: bool = True,
        f_filter: Callable[[T], bool] | None = None,
        sort: bool = True,
    ) -> Tree[T]:
        out = self.get_children(
            direct_only=True,
            types=types,
            f_filter=f_filter,
            sort=sort,
        )

        tree = Tree[T](
            {
                n: n.get_tree(
                    types=types,
                    include_root=False,
                    f_filter=f_filter,
                    sort=sort,
                )
                for n in out
            }
        )

        if include_root:
            if isinstance(self, types):
                if not f_filter or f_filter(self):
                    tree = Tree[T]({self: tree})

        return tree

    @staticmethod
    def get_nodes_from_gifs(gifs: Iterable[GraphInterface]):
        # TODO move this to gif?
        return {gif.node for gif in gifs}
        # TODO what is faster
        # return {n.node for n in gifs if isinstance(n, GraphInterfaceSelf)}

    # Hierarchy queries ----------------------------------------------------------------

    def get_hierarchy(self) -> list[tuple["Node", str]]:
        return [(cast_assert(Node, n), name) for n, name in super().get_hierarchy()]

    def get_parent_f(
        self,
        filter_expr: Callable[["Node"], bool],
        direct_only: bool = False,
        include_root: bool = True,
    ):
        parents = [p for p, _ in self.get_hierarchy()]
        if not include_root:
            parents = parents[:-1]
        if direct_only:
            parents = parents[-1:]
        for p in reversed(parents):
            if filter_expr(p):
                return p
        return None

    def get_parent_of_type[T: Node](
        self, parent_type: type[T], direct_only: bool = False, include_root: bool = True
    ) -> T | None:
        return cast(
            parent_type | None,
            self.get_parent_f(
                lambda p: isinstance(p, parent_type),
                direct_only=direct_only,
                include_root=include_root,
            ),
        )

    def get_parent_with_trait[TR: Trait](
        self, trait: type[TR], include_self: bool = True
    ):
        hierarchy = self.get_hierarchy()
        if not include_self:
            hierarchy = hierarchy[:-1]
        for parent, _ in reversed(hierarchy):
            if parent.has_trait(trait):
                return parent, parent.get_trait(trait)
        raise KeyErrorNotFound(f"No parent with trait {trait} found")

    def get_first_child_of_type[U: Node](self, child_type: type[U]) -> U:
        for level in self.get_tree(types=Node).iter_by_depth():
            for child in level:
                if isinstance(child, child_type):
                    return child
        raise KeyErrorNotFound(f"No child of type {child_type} found")

    # ----------------------------------------------------------------------------------
    def zip_children_by_name_with[N: Node](
        self, other: "Node", sub_type: type[N]
    ) -> dict[str, tuple[N, N]]:
        nodes = self, other
        children = tuple(
            Node.with_names(
                n.get_children(direct_only=True, include_root=False, types=sub_type)
            )
            for n in nodes
        )
        return zip_dicts_by_key(*children)

    @staticmethod
    def with_names[N: Node](nodes: Iterable[N]) -> dict[str, N]:
        return {n.get_name(): n for n in nodes}

    def __rich_repr__(self):
        if not self._is_setup:
            yield f"{type(self)}(not init)"
        else:
            yield self.get_full_name()

    __rich_repr__.angular = True

    def nearest_common_ancestor(self, *others: "Node") -> tuple["Node", str] | None:
        """
        Finds the nearest common ancestor of the given nodes, or None if no common
        ancestor exists
        """
        nodes = [self, *others]
        if not nodes:
            return None

        # Get hierarchies for all nodes
        hierarchies = [list(n.get_hierarchy()) for n in nodes]
        min_length = min(len(h) for h in hierarchies)

        # Find the last matching ancestor
        last_match = None
        for i in range(min_length):
            ref_node, ref_name = hierarchies[0][i]
            if any(h[i][0] is not ref_node for h in hierarchies[1:]):
                break
            last_match = (ref_node, ref_name)

        return last_match



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from typing import Mapping

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.core.trait import TraitImpl
from faebryk.libs.picker.picker import DescriptiveProperties


class has_descriptive_properties_defined(F.has_descriptive_properties.impl()):
    def __init__(
        self,
        properties: Mapping[str, str]
        | Mapping[DescriptiveProperties, str]
        | Mapping[str | DescriptiveProperties, str],
    ) -> None:
        super().__init__()
        self.properties: dict[str, str] = dict(properties)

    def get_properties(self) -> dict[str, str]:
        return self.properties

    def handle_duplicate(self, old: TraitImpl, node: Node) -> bool:
        if not isinstance(old, has_descriptive_properties_defined):
            assert isinstance(old, F.has_descriptive_properties)
            self.properties.update(old.get_properties())
            return super().handle_duplicate(old, node)

        old.properties.update(self.properties)
        old._handle_prop_set()
        return False

    def _handle_prop_set(self):
        obj = self.get_obj(type=Module)

        # TODO: deprecate using DescriptiveProperties directly and then get rid of this
        if (
            DescriptiveProperties.partno in self.properties
            and DescriptiveProperties.manufacturer in self.properties
        ):
            obj.add(
                F.is_pickable_by_part_number(
                    self.properties[DescriptiveProperties.manufacturer],
                    self.properties[DescriptiveProperties.partno],
                )
            )

        if "LCSC" in self.properties:
            obj.add(
                F.is_pickable_by_supplier_id(
                    self.properties["LCSC"],
                    supplier=F.is_pickable_by_supplier_id.Supplier.LCSC,
                )
            )

    def on_obj_set(self):
        super().on_obj_set()
        self._handle_prop_set()



"""
Datatypes used in the model.
"""

import logging
from contextlib import contextmanager
from typing import Iterable

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class Ref(tuple[str]):
    """Shell class to provide basic utils for a reference."""

    def add_name(self, name: str | int) -> "Ref":
        """Return a new Ref with the given name."""
        return Ref((*self, str(name)))

    def __str__(self) -> str:
        return ".".join(map(str, self))

    @classmethod
    def empty(cls) -> "Ref":
        """Return an empty Ref."""
        return cls(())

    @classmethod
    def from_one(cls, name: str | int) -> "Ref":
        """Return a Ref with a single item."""
        return cls((str(name),))


class KeyOptItem[V](tuple[Ref | None, V]):
    """A class representing anf optionally-named thing."""

    @property
    def ref(self) -> Ref | None:
        """Return the name of this item, if it has one."""
        return self[0]

    @property
    def value(self) -> V:
        """Return the value of this item."""
        return self[1]

    @classmethod
    def from_kv(cls, key: Ref | None, value: V) -> "KeyOptItem[V]":
        """Return a KeyOptItem with a single item."""
        return KeyOptItem((key, value))


class KeyOptMap[V](tuple[KeyOptItem[V]]):
    """A class representing a set of optionally-named things."""

    def keys(self) -> Iterable[Ref | None]:
        """Return an iterable of all the names in this set."""
        return map(lambda x: x.ref, filter(lambda x: x.ref is not None, self))

    def values(self) -> Iterable[V]:
        """Return an iterable of all the values in this set."""
        return map(lambda x: x.value, self)

    @classmethod
    def from_item(cls, item: KeyOptItem[V]) -> "KeyOptMap[V]":
        """Return a KeyOptMap with a single item."""
        return KeyOptMap((item,))

    @classmethod
    def from_kv(cls, key: Ref | None, value: V) -> "KeyOptMap[V]":
        """Return a KeyOptMap with a single item."""
        return cls.from_item(KeyOptItem.from_kv(key, value))

    @classmethod
    def empty(cls) -> "KeyOptMap[V]":
        """Return an empty KeyOptMap."""
        return cls(())


class StackList[V](list[V]):
    """Manages context while compiling ato code."""

    def __init__(self) -> None:
        super().__init__([])

    @contextmanager
    def enter(self, context: V):
        """Enter a context."""
        self.append(context)
        try:
            yield
        finally:
            self.pop()

    @property
    def top(self) -> V:
        """Return the current context."""
        return self[-1]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.moduleinterface import ModuleInterface


class Electrical(ModuleInterface):
    # potential= L.p_field(units=P.dimensionless)

    def get_net(self):
        from faebryk.library.Net import Net

        nets = {
            net
            for mif in self.get_connected()
            if (net := mif.get_parent_of_type(Net)) is not None
        }

        if not nets:
            return None

        assert len(nets) == 1
        return next(iter(nets))



"""
This module contains functions for interacting with layout data,
and generating files required to reuse layouts.

Thanks @nickkrstevski (https://github.com/nickkrstevski) for
the heavy lifting on this one!
"""

import copy
import hashlib
import json
import logging
import uuid
from pathlib import Path
from typing import TypedDict

from more_itertools import first

import faebryk.library._F as F
import faebryk.libs.exceptions
from atopile import errors, front_end
from atopile.address import AddressError, AddrStr
from atopile.config import ProjectConfig, config
from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.libs.util import (
    FuncDict,
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    find,
    get_module_from_path,
)

logger = logging.getLogger(__name__)


def _generate_uuid_from_string(path: str) -> str:
    """Spits out a uuid in hex from a string"""
    path_as_bytes = path.encode("utf-8")
    hashed_path = hashlib.blake2b(path_as_bytes, digest_size=16).digest()
    return str(uuid.UUID(bytes=hashed_path))


def _index_module_layouts() -> FuncDict[type[Module], set[Path]]:
    """Find, tag and return a set of all the modules with layouts."""
    directory = config.project.paths.root

    entries: FuncDict[type[Module], set[Path]] = FuncDict()
    ato_modules = front_end.bob.modules

    for filepath in directory.glob("**/ato.yaml"):
        with faebryk.libs.exceptions.downgrade(Exception, logger=logger):
            project_config = ProjectConfig.from_path(filepath.parent)

            if project_config is None:
                raise faebryk.libs.exceptions.UserResourceException(
                    f"Failed to load module config: {filepath}"
                )

            for build in project_config.builds.values():
                with faebryk.libs.exceptions.downgrade(Exception, logger=logger):
                    try:
                        entry_section = build.entry_section
                    except AddressError:
                        # skip builds with no entry, e.g. generics
                        # config validation happens before this point
                        continue

                    # Check if the module is a known python module
                    if (
                        class_ := get_module_from_path(
                            build.entry_file_path,
                            entry_section,
                            # we might have duplicates from different builds
                            allow_ambiguous=True,
                        )
                    ) is not None:
                        # we only bother to index things we've imported,
                        # otherwise we can be sure they weren't used
                        assert isinstance(class_, type)
                        entries.setdefault(class_, set()).add(build.paths.layout)

                    # Check if the module is a known ato module
                    elif class_ := ato_modules.get(
                        # This is the address w/ an absolute path to the entry file
                        # which is the format also used by the frontend to key modules
                        AddrStr.from_parts(build.entry_file_path, build.entry_section)
                    ):
                        entries.setdefault(class_, set()).add(build.paths.layout)

    return entries


class ModuleMap(TypedDict):
    layout_path: str
    """Absolute path to the KiCAD layout file"""

    uuid_to_addr_map: dict[str, str]
    """Map of UUIDs, as expected in the layout, to addresses"""

    addr_map: dict[str, str]
    """Map of addresses in the parent module to addresses in the child module"""

    group_components: list[str]
    """List of components in the group"""

    nested_groups: list[str]
    """List of addresses of nested groups"""


def generate_module_map(app: Module) -> None:
    """Generate a file containing a list of all the modules and their components in the build."""  # noqa: E501  # pre-existing
    module_map: dict[str, ModuleMap] = {}

    module_layouts = _index_module_layouts()

    for module, trait in GraphFunctions(app.get_graph()).nodes_with_trait(
        F.has_reference_layout
    ):
        assert isinstance(module, Module)
        module_layouts.setdefault(type(module), set()).update(trait.paths)

    def _dfs_layouts(module_instance: Module) -> tuple[list[str], set[str]]:
        """
        Pre-order DFS of the module's layout tree
        Returns:
            - a list of addresses of nested groups
            - a set of addresses of child components
        """
        nested_groups: list[str] = []
        child_component_addrs: set[str] = set()

        # FIXME: this was created for nested grouping, however it was too much effort
        # to support this in KiCAD, so we're ignoring it for now
        def _descend() -> None:
            for child_instance in module_instance.get_children_modules(
                types=Module, direct_only=True
            ):
                ngs, cca = _dfs_layouts(child_instance)
                nested_groups.extend(ngs)
                child_component_addrs.update(cca)

        try:
            # TODO: this could be improved if we had the mro of the module
            module_super = find(
                module_layouts.keys(), lambda x: isinstance(module_instance, x)
            )
        except KeyErrorNotFound:
            # For now, only descend if there isn't a layout for this module
            # This means we'll only get the top-level group
            _descend()

        except KeyErrorAmbiguous as e:
            raise errors.UserNotImplementedError(
                "There are multiple build configurations for this module.\n"
                "We don't currently support multiple layouts for the same module."
                "Show the issue some love to get it done: https://github.com/atopile/atopile/issues/399"
            ) from e

        else:
            # Build up a map of UUIDs of the children of the module
            # The keys are instance UUIDs and the values are the
            # corresponding UUIDs in the layout
            # Here we include everything, even if accounted for by a nested group
            uuid_map = {}
            addr_map = {}
            for inst_child in module_instance.get_children_modules(types=Module):
                if not inst_child.has_trait(F.has_footprint):
                    continue

                parent_addr = inst_child.get_full_name()
                # This is a bit of a hack that makes assumptions based on the addressing
                # scheme to get the child's address w/r/t to it's root module
                child_addr = inst_child.relative_address(module_instance)

                addr_map[parent_addr] = child_addr

                for addr in (parent_addr, child_addr):
                    uuid_map[_generate_uuid_from_string(addr)] = addr

            # Add the module map to the module map
            group_name = module_instance.relative_address(app)
            ungrouped_footprints = set(addr_map.keys()) - child_component_addrs

            module_map[group_name] = {
                "layout_path": str(first(module_layouts[module_super])),
                # This maps the UUIDs of footprints to their addresses
                # for compatibility with the old layouts
                # TODO: @v0.4 remove this
                "uuid_to_addr_map": uuid_map,
                "addr_map": addr_map,
                "group_components": list(ungrouped_footprints),
                "nested_groups": copy.copy(nested_groups),
            }

            # Reset the nested_groups here to add only next-of-kin
            nested_groups = [group_name]
            child_component_addrs.update(ungrouped_footprints)

        return nested_groups, child_component_addrs

    # Run the DFS on the root module
    # We add an additional loop at the top here to avoid creating a top-level group
    for child_instance in app.get_children_modules(types=Module, direct_only=True):
        _dfs_layouts(child_instance)

    # Dump the module map to the relevant layout file for the KiCAD extension to use
    with open(
        config.build.paths.output_base.with_suffix(".layouts.json"),
        "w",
        encoding="utf-8",
    ) as f:
        json.dump(module_map, f, indent=4)



from faebryk.libs.tools.typer import typer_callback
from faebryk.tools.project import main as project_main
from faebryk.tools.refactor import main as refactor_main


@typer_callback(None)
def main():
    print("Running faebryk")
    pass


def __main__():
    main.add_typer(project_main, name="project")
    main.add_typer(refactor_main, name="refactor")
    main()


if __name__ == "__main__":
    __main__()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Callable, cast

from faebryk.core.cpp import Graph
from faebryk.core.graph import GraphFunctions
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import NodeException
from faebryk.core.parameter import Expression, Parameter
from faebryk.libs.test.times import Times
from faebryk.libs.util import assert_once, cast_assert, find, groupby, not_none, once

logger = logging.getLogger(__name__)


class is_bus_parameter(Parameter.TraitT.decless()):
    """
    Marks a parameter as a bus parameter.
    This means that it's constraints are dependent on the bus connections.
    """

    type KEY = Callable[[ModuleInterface], Parameter] | Parameter
    type EXPR_FACTORY = Callable[[], Expression]
    type REDUCE = tuple[KEY, EXPR_FACTORY]

    def __init__(self, key: KEY | None = None, reduce: REDUCE | None = None):
        """
        args:
            key: name of the parameter to be reduced given as lambda: node.param
            reduce: operation and 'name' of source parameters to be reduced
        example:
            See ElectricPower.py
        """
        super().__init__()

        self._key = key
        self._reduce = reduce

        self._guard = False
        self._merged: set[int] = set()

    @staticmethod
    def _get_key_from_parameter(
        obj: Parameter,
    ) -> Callable[[ModuleInterface], Parameter]:
        # TODO consider handling fields as obj
        # e.g ElectricPower.max_current instead of self.max_current

        p = obj.get_parent()
        if not p:
            raise NodeException(obj, "Can't auto-detect key for non-parented parameter")
        parent, name = p
        if not isinstance(parent, ModuleInterface):
            raise NodeException(
                obj, "Can't auto-detect key for non-mif child parameter"
            )

        return lambda mif: getattr(mif, name)

    def on_obj_set(self):
        if isinstance(self._key, Callable):
            return

        if self._key is None:
            obj = self.get_obj(Parameter)
        else:
            obj = self._key

        self._key = self._get_key_from_parameter(obj)

    @once
    def mif_parent(self) -> ModuleInterface:
        return cast_assert(ModuleInterface, not_none(self.obj.get_parent())[0])

    @assert_once
    def resolve(self, mifs: set[ModuleInterface]):
        if self._guard:
            return

        assert isinstance(self._key, Callable)

        mif_parent = self.mif_parent()
        self_param = self.get_obj(Parameter)
        if self._key(mif_parent) is not self_param:
            raise NodeException(self, "Key not mapping to parameter")

        params = [self._key(mif) for mif in mifs]
        params_with_guard = [
            (param, param.get_trait(is_bus_parameter)) for param in params
        ]

        # Disable guards to prevent infinite recursion
        for param, guard in params_with_guard:
            guard._guard = True
            guard._merged.add(id(self_param))

        new_merge = {p for p in params if id(p) not in self._merged}

        # Alias bus parameters
        for param in new_merge:
            self._merged.add(id(param))
            self_param.alias_is(param)

        # Reduce
        if self._reduce:
            # TODO use new_merge, then we can get rid of assert_once
            source_key, reducer = self._reduce
            if not isinstance(source_key, Callable):
                source_key = self._get_key_from_parameter(source_key)
            source_params = [source_key(mif) for mif in mifs]
            self_param.alias_is(reducer(*source_params))

        # Enable guards again
        for _, guard in params_with_guard:
            guard._guard = False

    @staticmethod
    def resolve_bus_parameters(graph: Graph):
        bus_parameters = cast(
            list[tuple[Parameter, is_bus_parameter]],
            GraphFunctions(graph).nodes_with_trait(is_bus_parameter),
        )

        times = Times()

        busses: list[set[ModuleInterface]] = []

        params_grouped_by_mif = groupby(bus_parameters, lambda p: p[1].mif_parent())

        # find for all busses a mif that represents it, and puts its dynamic params here
        # we use the that connected mifs are the same type and thus have the same params
        # TODO: limitation: specialization (need to subgroup by type) (see exception)
        param_bus_representatives: set[tuple[Parameter, is_bus_parameter]] = set()

        while params_grouped_by_mif:
            bus_representative_mif, bus_representative_params = (
                params_grouped_by_mif.popitem()
            )
            # expensive call
            paths = bus_representative_mif.get_connected(include_self=True)
            connections = set(paths.keys())

            busses.append(connections)
            if len(set(map(type, connections))) > 1:
                raise NotImplementedError(
                    "No support for specialized bus with dynamic params"
                )

            for m in connections:
                if m in params_grouped_by_mif:
                    del params_grouped_by_mif[m]
            param_bus_representatives |= set(bus_representative_params)

        times.add("get parameter connections")

        # exec resolution
        for _, trait in param_bus_representatives:
            bus_representative_mif = trait.mif_parent()
            param_bus = find(busses, lambda bus: bus_representative_mif in bus)
            trait.resolve(param_bus)

        times.add("merge parameters")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(times)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class App(Module):
    led: F.PoweredLED
    battery: F.Battery

    def __preinit__(self) -> None:
        self.led.power.connect(self.battery.power)

        self.led.led.add(F.has_explicit_part.by_supplier("C965802"))
        self.led.led.forward_voltage.alias_is(2.4 * P.V)
        self.led.led.max_brightness.alias_is(435 * P.millicandela)
        self.led.led.max_current.alias_is(20 * P.mA)
        self.led.led.color.alias_is(F.LED.Color.YELLOW)

        self.led.current_limiting_resistor.add(
            F.has_explicit_part.by_supplier("C159037")
        )

        buttoncell = self.battery.specialize(F.ButtonCell())
        buttoncell.add(
            F.has_explicit_part.by_supplier(
                "C5239862",
                pinmap={
                    "1": self.battery.power.lv,
                    "2": self.battery.power.hv,
                },
            )
        )
        buttoncell.voltage.alias_is(3 * P.V)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
from faebryk.exporters.pcb.layout.heuristic_decoupling import (
    LayoutHeuristicElectricalClosenessDecouplingCaps,
)
from faebryk.exporters.pcb.layout.heuristic_pulls import (
    LayoutHeuristicElectricalClosenessPullResistors,
)
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class App(Module):
    leds: F.PoweredLED
    battery: F.Battery
    eeprom: F.M24C08_FMN6TP

    def __preinit__(self) -> None:
        self.leds.power.connect(self.battery.power)
        self.battery.power.voltage.alias_is(3 * P.V)

        # Parametrize
        self.leds.led.color.constrain_subset(F.LED.Color.YELLOW)
        self.leds.led.brightness.constrain_subset(
            TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
        )

        self.eeprom.ic.power.voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )
        self.eeprom.ic.set_address(0x0)
        self.eeprom.ic.data.pull_up_sda.resistance.constrain_subset(
            L.Range.from_center_rel(2 * P.kohm, 0.1)
        )
        self.eeprom.ic.data.pull_up_scl.resistance.constrain_subset(
            L.Range.from_center_rel(2 * P.kohm, 0.1)
        )

        # Layout
        Point = F.has_pcb_position.Point
        Ly = F.has_pcb_position.layer_type

        layout = LayoutTypeHierarchy(
            layouts=[
                LayoutTypeHierarchy.Level(
                    mod_type=F.PoweredLED,
                    layout=LayoutAbsolute(Point((0, 0, 0, Ly.TOP_LAYER))),
                    children_layout=LayoutTypeHierarchy(
                        layouts=[
                            LayoutTypeHierarchy.Level(
                                mod_type=(F.LED, F.Resistor),
                                layout=LayoutExtrude((0, 5)),
                            ),
                        ]
                    ),
                ),
                LayoutTypeHierarchy.Level(
                    mod_type=F.Battery,
                    layout=LayoutAbsolute(Point((0, 20, 0, Ly.BOTTOM_LAYER))),
                ),
                LayoutTypeHierarchy.Level(
                    mod_type=F.M24C08_FMN6TP,
                    layout=LayoutAbsolute(Point((15, 10, 0, Ly.TOP_LAYER))),
                ),
            ]
        )
        self.add(F.has_pcb_layout_defined(layout))
        self.add(F.has_pcb_position_defined(Point((50, 50, 0, Ly.NONE))))

        LayoutHeuristicElectricalClosenessDecouplingCaps.add_to_all_suitable_modules(
            self
        )
        LayoutHeuristicElectricalClosenessPullResistors.add_to_all_suitable_modules(
            self
        )

        # TODO: remove when we have a LED picker
        self.leds.led.add(F.has_explicit_part.by_supplier("C965802"))
        self.leds.led.forward_voltage.alias_is(2.4 * P.V)
        self.leds.led.max_brightness.alias_is(435 * P.millicandela)
        self.leds.led.max_current.alias_is(20 * P.mA)
        self.leds.led.color.alias_is(F.LED.Color.YELLOW)

        # TODO remove when we have a battery picker
        self.battery.add(
            F.has_explicit_part.by_supplier(
                "C5239862",
                pinmap={
                    "1": self.battery.power.lv,
                    "2": self.battery.power.hv,
                },
            )
        )
        self.battery.voltage.alias_is(3 * P.V)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import subprocess as sp
import tempfile
from pathlib import Path
from zipfile import ZipFile

from kicadcliwrapper.generated.kicad_cli import kicad_cli as k

logger = logging.getLogger(__name__)


def _export(cmd):
    return k(k.pcb(k.pcb.export(cmd))).exec()


def export_step(pcb_file: Path, step_file: Path) -> None:
    """
    3D PCBA STEP file export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.step(
                INPUT_FILE=str(pcb_file),
                force=True,
                no_dnp=True,
                subst_models=True,
                output=str(step_file.absolute()),
            )
        )
    except sp.CalledProcessError as e:
        raise Exception("Failed to export step file") from e


def export_dxf(pcb_file: Path, dxf_file: Path) -> None:
    """
    PCB outline export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.dxf(
                INPUT_FILE=str(pcb_file),
                exclude_refdes=True,
                exclude_value=True,
                output_units="mm",
                layers="Edge.Cuts",
                output=str(dxf_file),
            )
        )
    except sp.CalledProcessError as e:
        raise Exception("Failed to export dxf file") from e


def export_glb(pcb_file: Path, glb_file: Path) -> None:
    """
    3D PCBA GLB file export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.glb(
                INPUT_FILE=str(pcb_file),
                force=True,
                include_tracks=True,
                include_zones=True,
                grid_origin=True,
                subst_models=True,
                no_dnp=True,
                output=str(glb_file.absolute()),
            )
        )
    except sp.CalledProcessError as e:
        raise Exception("Failed to export glb file") from e


def export_svg(pcb_file: Path, svg_file: Path, flip_board: bool = False) -> None:
    """
    2D PCBA SVG file export using the kicad-cli
    """

    layers = "F.Cu,F.Paste,F.SilkS,F.Mask,Edge.Cuts"
    if flip_board:
        layers = layers.replace("F.", "B.")

    try:
        _export(
            k.pcb.export.svg(
                INPUT_FILE=str(pcb_file),
                layers=layers,
                page_size_mode="2",
                exclude_drawing_sheet=True,
                output=str(svg_file),
            )
        )
    except sp.CalledProcessError as e:
        raise Exception("Failed to export svg file") from e


def export_gerber(pcb_file: Path, gerber_zip_file: Path) -> None:
    """
    Gerber export using the kicad-cli
    """

    logger.info(f"Exporting gerber files to {gerber_zip_file}")
    gerber_dir = gerber_zip_file.parent
    gerber_dir.mkdir(parents=True, exist_ok=True)

    # Create a temporary folder to export the gerber and drill files to
    with tempfile.TemporaryDirectory(dir=gerber_dir) as temp_dir:
        out_dir = temp_dir if temp_dir.endswith("/") else f"{temp_dir}/"

        try:
            _export(
                k.pcb.export.gerbers(
                    INPUT_FILE=str(pcb_file),
                    layers="F.Cu,B.Cu,F.Paste,B.Paste,F.SilkS,B.SilkS,F.Mask,B.Mask,F.CrtYd,B.CrtYd,F.Fab,B.Fab,Edge.Cuts",
                    output=out_dir,
                )
            )
        except sp.CalledProcessError as e:
            raise Exception("Failed to export gerber files") from e

        try:
            _export(
                k.pcb.export.drill(
                    INPUT_FILE=str(pcb_file),
                    format="excellon",
                    excellon_separate_th=True,
                    generate_map=True,
                    map_format="gerberx2",
                    output=out_dir,
                )
            )

        except sp.CalledProcessError as e:
            raise Exception("Failed to export drill files") from e

        # Zip the gerber files
        with ZipFile(gerber_zip_file, "w") as zipf:
            for file in Path(temp_dir).iterdir():
                if file.is_file():
                    zipf.write(file, arcname=file.name)


def export_pick_and_place(pcb_file: Path, pick_and_place_file: Path) -> None:
    """
    Pick and place export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.pos(
                INPUT_FILE=str(pcb_file),
                side="both",
                format="csv",
                units="mm",
                exclude_dnp=True,
                output=str(pick_and_place_file),
            )
        )
    except sp.CalledProcessError as e:
        raise Exception("Failed to export pick and place file") from e



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import override

import faebryk.library._F as F
from faebryk.core.module import Module


class has_explicit_part(Module.TraitT.decless()):
    mfr: str
    partno: str
    supplier_id: str
    supplier_partno: str
    pinmap: dict[str, F.Electrical | None] | None
    override_footprint: tuple[F.Footprint, str] | None = None

    @classmethod
    def by_mfr(
        cls,
        mfr: str,
        partno: str,
        pinmap: dict[str, F.Electrical | None] | None = None,
        override_footprint: tuple[F.Footprint, str] | None = None,
    ):
        out = cls()
        out.mfr = mfr
        out.partno = partno
        out.pinmap = pinmap
        out.override_footprint = override_footprint
        return out

    @classmethod
    def by_supplier(
        cls,
        supplier_partno: str,
        supplier_id: str = "lcsc",
        pinmap: dict[str, F.Electrical | None] | None = None,
        override_footprint: tuple[F.Footprint, str] | None = None,
    ):
        if supplier_id != "lcsc":
            raise NotImplementedError(f"Supplier {supplier_id} not supported")
        out = cls()
        out.supplier_id = supplier_id
        out.supplier_partno = supplier_partno
        out.pinmap = pinmap
        out.override_footprint = override_footprint
        return out

    @override
    def on_obj_set(self):
        super().on_obj_set()
        obj = self.get_obj(type=Module)

        if hasattr(self, "mfr"):
            assert self.partno
            obj.add(F.is_pickable_by_part_number(self.mfr, self.partno))
        if hasattr(self, "supplier_partno"):
            assert self.supplier_id == "lcsc"
            obj.add(
                F.is_pickable_by_supplier_id(
                    self.supplier_partno,
                    supplier=F.is_pickable_by_supplier_id.Supplier.LCSC,
                )
            )

        if self.pinmap:
            obj.add(F.can_attach_to_footprint_via_pinmap(self.pinmap))

        if self.override_footprint:
            fp, fp_kicad_id = self.override_footprint
            fp.add(F.KicadFootprint.has_kicad_identifier(fp_kicad_id))
            obj.get_trait(F.can_attach_to_footprint).attach(fp)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
import subprocess
import tempfile
from pathlib import Path

import typer

logger = logging.getLogger(__name__)
app = typer.Typer()


def is_running_in_vscode_terminal() -> bool:
    return (
        "TERM_PROGRAM" in os.environ and "vscode" in os.environ["TERM_PROGRAM"].lower()
    )


def get_vscode_instances_count(binname: str) -> int:
    try:
        result = subprocess.run(["ps", "aux"], capture_output=True, text=True)
        return len(
            [
                line
                for line in result.stdout.splitlines()
                if f"/{binname}" in line and "grep" not in line
            ]
        )
    except subprocess.SubprocessError:
        return 0


def get_code_binary() -> Path | None:
    candidates = ["cursor", "code"]
    for candidate in candidates:
        result = subprocess.run(["which", candidate], capture_output=True, text=True)
        if result.returncode == 0:
            return Path(result.stdout.strip())
    return None


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def cprofile(
    ctx: typer.Context, snakeviz: bool = typer.Option(False, help="Use snakeviz")
):
    """Profile a Python program using cProfile."""
    if not ctx.args:
        typer.echo("No command provided to profile", err=True)
        raise typer.Exit(1)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        pstats_file = temp_dir_path / "output.pstats"
        dot_file = temp_dir_path / "output.dot"

        # Run cProfile
        subprocess.run(
            ["python", "-m", "cProfile", "-o", str(pstats_file), *ctx.args], check=True
        )

        # Convert to dot format
        subprocess.run(
            ["gprof2dot", "-f", "pstats", str(pstats_file), "-o", str(dot_file)],
            check=True,
        )

        # Determine how to display the output
        code_bin = get_code_binary()

        if is_running_in_vscode_terminal():
            subprocess.run(["code", str(dot_file)], check=True)
        elif code_bin and get_vscode_instances_count(code_bin.name) > 0:
            subprocess.run([str(code_bin), "-r", str(dot_file)], check=True)
        else:
            png_file = temp_dir_path / "output.png"
            subprocess.run(
                ["dot", "-Tpng", "-o", str(png_file), str(dot_file)], check=True
            )
            subprocess.Popen(["xdg-open", str(png_file)])

        if snakeviz:
            subprocess.run(["snakeviz", str(pstats_file)], check=True)

        # Display with cursor
        typer.echo(str(dot_file))


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def viztracer(ctx: typer.Context):
    """Profile a Python program using VizTracer."""
    if not ctx.args:
        typer.echo("No command provided to profile", err=True)
        raise typer.Exit(1)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        json_file = temp_dir_path / "output.json"
        subprocess.run(
            ["viztracer", "--open", "-o", str(json_file), *ctx.args], check=True
        )


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def pyinstrument(ctx: typer.Context):
    """Profile a Python program using Pyinstrument."""
    if not ctx.args:
        typer.echo("No command provided to profile", err=True)
        raise typer.Exit(1)

    subprocess.run(["pyinstrument", "-r", "text", *ctx.args], check=True)


if __name__ == "__main__":
    app()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import TYPE_CHECKING, Sequence

import faebryk.library._F as F
from faebryk.core.node import Node
from faebryk.exporters.pcb.routing.util import (
    Path,
    Route,
    get_internal_nets_of_node,
    get_pads_pos_of_mifs,
)

if TYPE_CHECKING:
    from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer

logger = logging.getLogger(__name__)


class has_pcb_routing_strategy_manual(F.has_pcb_routing_strategy.impl()):
    def __init__(
        self,
        paths: Sequence[tuple[F.Net | Sequence[F.Electrical], Path]],
        relative_to: Node | None = None,
        absolute: bool = False,
    ):
        super().__init__()

        self.paths_rel = paths
        self.relative_to = relative_to
        self.absolute = absolute

    def calculate(self, transformer: "PCB_Transformer"):
        node = self.obj
        nets = get_internal_nets_of_node(node)

        relative_to = (self.relative_to or self.obj) if not self.absolute else None

        if relative_to:
            pos = relative_to.get_parent_with_trait(F.has_pcb_position)[
                1
            ].get_position()
            for _, path in self.paths_rel:
                path.abs_pos(pos)

        def get_route_for_mifs_in_net(mifs, path):
            pads = get_pads_pos_of_mifs(mifs)
            # Nothing we have can do with the groups because we don't know
            #  which parts of the path are specifically for those pads
            # pad_groups = group_pads_that_are_connected_already(pads)

            return Route(pads=pads, path=path)

        return [
            route
            for net_or_mifs, path in self.paths_rel
            if (
                route := get_route_for_mifs_in_net(
                    nets[net_or_mifs]
                    if isinstance(net_or_mifs, F.Net)
                    else net_or_mifs,
                    path,
                )
            )
        ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class WCHJiangsu_Qin_Heng_CH224K_ReferenceDesign(Module):
    """
    USB PD and Other Fast Charging Protocol Sink Controller Reference Design
    Design is ment for connecting to a female USB Type-C connector.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    controller: F.WCHJiangsu_Qin_Heng_CH224K

    vbus: F.ElectricPower
    usb_data: F.USB2_0_IF.Data
    cc = L.list_field(2, F.Electrical)
    vdd_resistor: F.Resistor

    power_good_indicator = L.f_field(F.LEDIndicator)(use_mosfet=False, active_low=True)

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.controller.vbus_detect.set_weak(
            on=True, owner=self
        ).resistance.constrain_subset(L.Range.from_center_rel(10 * P.kohm, 0.01))
        self.controller.vbus_detect.reference.connect(self.vbus)

        self.controller.power.hv.connect_via(self.vdd_resistor, self.vbus.hv)
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)

        self.controller.cc1.line.connect(self.cc[0])
        self.controller.cc2.line.connect(self.cc[1])
        self.controller.usb.connect(self.usb_data)

        self.power_good_indicator.logic_in.connect(self.controller.power_good)
        self.power_good_indicator.logic_in.reference.connect(self.vbus)

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.controller.power.decoupled.decouple(self).capacitance.constrain_subset(
            L.Range.from_center_rel(1 * P.uF, 0.1)
        )

        self.vdd_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1 * P.kohm, 0.01)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module

logger = logging.getLogger(__name__)


class Net(Module):
    part_of: F.Electrical

    def get_connected_pads(self) -> dict[F.Pad, F.Footprint]:
        """Return a dict of pads connected to this net"""
        return {
            pad: fp
            for mif in self.get_connected_interfaces()
            if (fp := mif.get_parent_of_type(F.Footprint)) is not None
            and (pad := mif.get_parent_of_type(F.Pad)) is not None
        }

    # TODO should this be here?
    def get_connected_interfaces(self):
        return {
            mif
            for mif in self.part_of.get_connected()
            # TODO: this should be removable since,
            # only mifs of the same type can connect
            if isinstance(mif, type(self.part_of))
        }

    def __repr__(self) -> str:
        up = super().__repr__()
        if self.has_trait(F.has_overriden_name):
            return f"{up}'{self.get_trait(F.has_overriden_name).get_name()}'"
        else:
            return up

    @classmethod
    def with_name(cls, name: str) -> "Net":
        n = cls()
        n.add(F.has_overriden_name_defined(name))
        return n

    @classmethod
    def find_from_part_of_mif(cls, mif: F.Electrical) -> "Net | None":
        """Return the Net that this "part_of" mif represents"""
        parent = mif.get_parent()
        if parent is None:
            return None
        if isinstance(parent[0], cls):
            return parent[0]
        return None
